%{
/**********************************************************************
Cpal.lex file, source file for flex scanner generator.
Main function located at the end of this file.
**********************************************************************/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include "y.tab.h"
#include "include.h"
#include "generate.h"

  extern int yydebug;
  extern int yyparse();
  extern rettype yylval;

  unsigned linenum=1;

  void parsestr(char yy[],char **st)
    {
      int idx,idx2=0,len;
      int quot=0;
      char *ret;
      len=strlen(yy);

      ret=(char *)malloc(len+1);
      for(idx=1;idx<len-1;idx++)
	{
	  if ((quot==0) && (yy[idx]=='"'))
	    quot=1;
	  else if ((quot==1) && (yy[idx]=='"'))
	    {
	      ret[idx2++]=yy[idx];
	      quot=0;
	    }
	  else if (quot==1)
	    {
	      printf("MAJOR ERROR!  SHOULD NOT RUN THIS LINE\n");
	      quot=0;
	    }
	  else
	    ret[idx2++]=yy[idx];
	}
      ret[idx2]=0;
      *st=ret;
    } 

%}
DIGIT [0-9]
ID	[A-Za-z][A-Za-z0-9]*
%%
"/*"                                            {
                                                  int c=0,c2=0;
                                                  do {
                                                    c2=c;
					            c=yyinput();
						    if (c=='\n') linenum++;
					        } while ((c!='/' || c2!='*') && (c!=EOF));
						  if (c==EOF){ printf("Scan ERROR Missing terminating */\n");
						  return ERROR;}
						  /*blank space printf("CBLANK(%s)",yytext);*/
                                                }
var                                             {
						yylval.lineno=linenum;
						return VAR;
                                                }
int                                             {
						yylval.lineno=linenum;
                                                return INT;
                                                }
real                                            {
						yylval.lineno=linenum;
                                                return REAL;
                                                }
string                                          {
						yylval.lineno=linenum;
                                                return STRING;
                                                }
boole                                           {
						yylval.lineno=linenum;
                                                return BOOLE;
                                                }
array                                           {
						yylval.lineno=linenum;
                                                return ARRAY;
                                                }
of                                              {
						yylval.lineno=linenum;
                                                return OF;
                                                }
begin                                           {
						yylval.lineno=linenum;
                                                return BEGN;
                                                }
end                                             {
						yylval.lineno=linenum;
                                                return END;
                                                }
read                                            {
						yylval.lineno=linenum;
                                                return READ;
                                                }
write                                           {
						yylval.lineno=linenum;
                                                return WRITE;
                                                }
if                                              {
						yylval.lineno=linenum;
                                                return IF;
                                                }
then                                            {
						yylval.lineno=linenum;
                                                return THEN;
                                                }
else                                            {
						yylval.lineno=linenum;
                                                return ELSE;
                                                }
while                                           {
						yylval.lineno=linenum;
                                                return WHILE;
                                                }
do                                              {
						yylval.lineno=linenum;
                                                return DO;
                                                }
return                                          {
						yylval.lineno=linenum;
                                                return RETURN;
                                                }
{DIGIT}+					{
                                                yylval.u.intval=atoi(yytext);
						yylval.lineno=linenum;
                                                /*printf("<INTEGER LITERAL: %d>\n",atoi(yytext));*/
                                                return NUM;
       						}
{DIGIT}*(("."{DIGIT}*("e"("+"|"-")?)?{DIGIT}+)|({DIGIT}+"e"("+"|"-")?{DIGIT}+)){DIGIT}*         {
                                                yylval.u.realval=atof(yytext);
						yylval.lineno=linenum;
                                                return FLOAT;
						/*printf("<REAL LITERAL: %s>\n",yytext);*/
						}
{DIGIT}+"."/[^0-9"."]                           {
                                                yylval.u.realval=atof(yytext);
						yylval.lineno=linenum;
                                                return FLOAT;
                                                /*printf("<REAL LITERAL: %s>\n",yytext);*/
                                                }
"!"                                             {
						yylval.lineno=linenum;
                                                return OPNOT;
                                                }
"*"                                             {
						yylval.lineno=linenum;
                                                return OPMULT; 
                                                }
"/"                                             {
						yylval.lineno=linenum;
                                                return OPDIV;
                                                }
"%"                                             {
						yylval.lineno=linenum;
                                                return OPMOD;
                                                }
"+"                                             {
						yylval.lineno=linenum;
                                                return OPADD;
                                                }
"-"                                             {
						yylval.lineno=linenum;
                                                return OPSUB;
                                                }
"="                                             {
						yylval.lineno=linenum;
                                                return OPEQ;
                                                }
"<"                                             {
						yylval.lineno=linenum;
                                                return OPLT;
                                                }
"<="                                            {
						yylval.lineno=linenum;
                                                return OPLTE;
                                                }
">"                                             {
						yylval.lineno=linenum;
                                                return OPGT;
                                                }
">="                                            {
						yylval.lineno=linenum;
                                                return OPGTE;
                                                }
"<>"                                            {
						yylval.lineno=linenum;
                                                return OPNEQ;
                                                }
"||"                                            {
						yylval.lineno=linenum;
                                                return OPOR;
                                                }
"&&"                                            {
						yylval.lineno=linenum;
                                                return OPAND;
                                                }
":="                                            {
						yylval.lineno=linenum;
                                                return OPSET;
                                                }
"["                                             {
						yylval.lineno=linenum;
                                                return SEPLBK;
                                                }
"]"                                             {
						yylval.lineno=linenum;
                                                return SEPRBK;
                                                }
"("                                             {
						yylval.lineno=linenum;
                                                return SEPLPAR;
                                                }
")"                                             {
						yylval.lineno=linenum;
                                                return SEPRPAR;
                                                }
":"                                             {
						yylval.lineno=linenum;
                                                return SEPCOL;
                                                }
";"                                             {
						yylval.lineno=linenum;
                                                return SEPSEMI;
                                                }
","                                             {
						yylval.lineno=linenum;
                                                return SEPCOM;
                                                }
".."                                            {
						yylval.lineno=linenum;
                                                return SEPDD;
                                                }
{ID}+                            		{
                                                yylval.u.strval=(char *)malloc(strlen(yytext)+1);
                                                strcpy(yylval.u.strval,yytext);
						yylval.lineno=linenum;
                                                return ID;
						}
\"([^\"\n]*(\"\")*[^\"\n]*)*\"                  {
                                                char *n;
                                                parsestr(yytext,&n);
						/*printf("%s\n",n);*/
                                                yylval.u.strval=n;
						yylval.lineno=linenum;
                                                return STRLIT;
                                                }
\"[^\"]*\n                                      {
                                                printf("Scan ERROR Missing Terminating \" on line %d\n",linenum);
						linenum++;
						return ERROR;
                                                }
(" "|\t)+                                       {
                                                /*blank space*/
                                                }
(\n)                                            linenum++;
.                                               {
                                                printf("Scan ERROR: %c on line %d\n",yytext,linenum);
						return ERROR;
                                                }
%%

void yyerror(const char *msg)
{
  fprintf(yyout,"Syntax Error on line [%d] with token [%s]\n",linenum, yytext);
}
int GetToken()
{
  return yylex();
}

int main(int argc, char *argv[])
{
  FILE *fin;
  char *outfn;
  int x;
  yydebug=1;
  if (argc!=2)
    {
      printf("Usage:  %s inputfilename \nmake",argv[0]);
      return 0;
    }
  fin=fopen(argv[1],"r");
  openfile(argv[1], &fout, &outfn);
  yyin=fin;
  x=yyparse();
  /*  printf("yyparse returns: [%d] on line [%d].\n",x,linenum);*/
  if ((x==0) && (globerr==0))
    printf("Compilation of [%s] Successful!\n",outfn);
  else
    {
      if(x!=0)
	printf("Parser Error exiting from line [%d]\n",linenum);
      if (globerr==-1)
	printf("Semantic Error!\n");
      printf("Compilation Failed!\n");
      remove(outfn);
    }
  fclose(yyin);
  free(outfn);
  return 0;
}




