/*********************************************************
funcs.h function header file

This header file includes functions to traverse the
Abstract Syntax Tree and to Print it as a cpal source.

Pretty printing feature can be turned off by setting the
#define blocksize 0

Pretty printint looks best with
#define blocksize 2
*********************************************************/

#ifndef funcs_h
#define funcs_h

#include "include.h"
#include "tab.h"
#include "semantic.h"

int lastnum=0;
int block=0;

/**********Set pretty printing indentation size**********/
#define blocksize 2


/*********************************************************
Miscellaneous functions
*********************************************************/

void quotparse(char *st, char **s)
{
  char *tmp;
  int i,i2,quote;
  i=0;
  i2=0;
  quote=0;

  while ((st)[i]!=0)
    {
      if ((st)[i]=='\"') quote++;
      i++;
    }
  tmp=(char *)malloc(i+1+quote);
  i=0;
  while ((st)[i]!=0)
    {
      tmp[i2++]=(st)[i++];
      if (tmp[i2-1]=='\"')
	{
	  tmp[i2++]='\"';
	}
    }
  tmp[i2]=0;
  *s=tmp;
  return;
}

void freelist(A_node lst)
{
  A_node idx,last;
  idx=lst;
  while (idx!=NULL)
    {
      last=idx->next;
      free(idx);
      idx=last;
    }
  return;
}

void prtlist(A_node lst)
{
  /*  A_node idx,last;
  idx=lst;
  while (idx!=NULL)
    {
      printf("%s, ",idx->u.decl.id);
      idx=idx->next;
    }
  printf("\n");
  return;*/
}

void blocksp()   /*This function does the pretty printing*/
{
  int i;
  for (i=0;i<block*blocksize;i++)
    printf(" ");
}

void linebreak(A_node prg,int flag=0) /*This function synchronizes lines*/
{
  if (flag==0) return;
  while(lastnum!=prg->lineno)
    {
      printf("\n");
      blocksp();
      /*printf("[%d]:",lastnum);/*Special Quick unComment Code*/
     
      if (lastnum>prg->lineno)
	lastnum=prg->lineno-1;
      lastnum++;
    }
}

/**********************************************************
Forward declaration stubs
**********************************************************/

void prtdecl(A_node, int=0);
void prttyp(A_node, int=0);
void prtstmt(A_node, int=0);
void prtexp(A_node, int=0);
void prtoper(A_node, int=0);
void prtvar(A_node, int=0);
void prtid(A_node, int=0);
void prtidx(A_node);
void travdecl(A_node, int=0, A_node *list=NULL);
int travtyp(A_node, int=0);
int travstmt(A_node, int=0, int=0, rettype **val=NULL);
int travexp(A_node, int, rettype **, int *dim=NULL, int **range=NULL, int =-1);
int travoper(A_node, int=0, rettype **val=NULL, int *dim=NULL, int **range=NULL);
int travvar(A_node, int=0, rettype **val=NULL, int *dim=NULL, int **range=NULL);
void travid(A_node, int=0, A_node *list=NULL, int=0, A_node=NULL,rettype *val =NULL);
void travidx(A_node, int=0, int =0, char *id=NULL);
void subcomplete(char *, A_node, A_node);

 /*************************************************************
 Abstract Syntax Tree Printing Functions
 *************************************************************/
 void prtprg(A_node prg, int mode=0) /*Main recursive loopback*/
 {
   if (prg==NULL) return;
   linebreak(prg);

   switch (prg->entity)
     {
     case declTag:
       prtdecl(prg,mode);
       break;
     case typTag:
       prttyp(prg,mode);
       break;
     case stmtTag:
       prtstmt(prg,mode);
       break;
     case expTag:
       prtexp(prg,mode);
       break;
     case operTag:
       prtoper(prg,mode);
       break;
     case varTag:
       prtvar(prg,mode);
       break;
     case idTag:
       prtid(prg,mode);
       break;
     case undefTag:
     default:
       printf("Error on line[%d]! [%d]\n",prg->lineno,prg->entity);
       globerr=-1;
       return;
       break;
     }

   return;
 }

 void prtdecl(A_node prg, int mode)
 {
   char *st;
   linebreak(prg,1);
   switch (prg->u.decl.kind)
     {
     case progDecl:
       printf("%s; ",prg->u.decl.id);
       block++;
       prtprg(prg->u.decl.u.prog.declist);
       prtprg(prg->u.decl.u.prog.subdecl);
       prtprg(prg->u.decl.u.prog.stmtblk);
       block--;
       blocksp();
       printf("\nend %s",prg->u.decl.id);
       lastnum++;
       break;
     case varDecl:
       if (mode==0)
	 {
	   printf("var ");
	 }
       prtprg(prg->u.decl.u.var.idList);
       printf(": ");
       prtprg(prg->u.decl.u.var.typ);
       if (mode==0)
	 {
	   printf("; ");
	 }
       if (prg->u.decl.next!=NULL)
	 {
	   if (mode==1)
	     {
	       printf("; ");
	     }
	   prtprg(prg->u.decl.next,mode);
	 }
       break;
     case constDecl:
       if (mode==0)
	 {
	   printf("var ");
	 }
       prtprg(prg->u.decl.u.cnst.idList);
       printf(": ");
       switch(prg->u.decl.u.cnst.kind)
	 {
	 case realtyp:
	   printf("%f",prg->u.decl.u.cnst.val.realval);
	   break;
	 case inttyp:
	   printf("%d",prg->u.decl.u.cnst.val.intval);
	   break;
	 case strtyp:
	   quotparse(prg->u.decl.u.cnst.val.strval,&st);
	   printf("\"%s\"",st);
	   free(st);
	   break;
	 case subrealtyp:
	   printf("-%f",prg->u.decl.u.cnst.val.realval);
	   break;
	 case subinttyp:
	   printf("-%d",prg->u.decl.u.cnst.val.intval);
	   break;
	 case addrealtyp:
	   printf("+%f",prg->u.decl.u.cnst.val.realval);
	   break;
	 case addinttyp:
	   printf("+%d",prg->u.decl.u.cnst.val.intval);
	   break;
	 case consttyp:
	   printf("%s",prg->u.decl.id);
	   break;
	 }
       if (mode==0)
	 {
	   printf("; ");
	 }
       if (prg->u.decl.next!=NULL)
	 {
	   if (mode==1)
	     {
	       printf("; ");
	     }
	   prtprg(prg->u.decl.next,mode);
	 }
       break;
     case subroutineDecl:
       printf("%s (",prg->u.decl.id);
       prtprg(prg->u.decl.u.sub.vallist,1);
       if (prg->u.decl.u.sub.reflist!=NULL)
	 {      
	   if (prg->u.decl.u.sub.vallist!=NULL)
	     printf("; ");
	   printf("var ");
	   prtprg(prg->u.decl.u.sub.reflist,1);
	 }
       printf(")");
       if (prg->u.decl.u.sub.rettype!=NULL)      
	 if (prg->u.decl.u.sub.rettype->u.typ.base!=0)
	   {
	     printf(": ");
	     prtprg(prg->u.decl.u.sub.rettype);
	   }
       printf("; ");
       block++;
       prtprg(prg->u.decl.u.sub.stmtblk);
       block--;
       printf("\n");
       blocksp();
       printf("end %s",prg->u.decl.id);
       lastnum++;
       prtprg(prg->u.decl.u.sub.next);
       break;
     }
 }

 void prttyp(A_node prg, int mode)
 {
   linebreak(prg,1);
   switch(prg->u.typ.kind)
     {
     case baseTy:
       switch(prg->u.typ.base)
	 {
	 case intTy:
	   printf("int");
	   break;
	 case realTy:
	   printf("real");
	   break;
	 case strTy:
	   printf("string");
	   break;
	 case boolTy:
	   printf("boole");
	   break;
	 }
       break;
     case array1Ty:
       if (prg->u.typ.base!=0)
	 printf("array [");
       switch(prg->u.typ.base1)
	 {
	 case inttyp:
	   printf("%d",prg->u.typ.r1.num);
	   break;
	 case subinttyp:
	   printf("-%d",prg->u.typ.r1.num);	
	   break;
	 case consttyp:
	   printf("%s",prg->u.typ.r1.id);
	   break;
	 }
       if (prg->next!=NULL)
	 {
	   printf(", ");
	   prtprg(prg->next);
	 }
       switch(prg->u.typ.base)
	 {
	 case intTy:
	   printf("] of int");
	   break;
	 case realTy:
	   printf("] of real");
	   break;
	 case strTy:
	   printf("] of string");
	   break;
	 case boolTy:
	   printf("] of boole");
	   break;
	 }
       break;
     case array2Ty:
       if (prg->u.typ.base!=0)
	 printf("array [");
       switch(prg->u.typ.base1)
	 {
	 case inttyp:
	   printf("%d",prg->u.typ.r1.num);
	   break;
	 case subinttyp:
	   printf("-%d",prg->u.typ.r1.num);	
	   break;
	 case consttyp:
	   printf("%s",prg->u.typ.r1.id);
	   break;
	 }
       printf("..");
       switch(prg->u.typ.base2)
	 {
	 case inttyp:
	   printf("%d",prg->u.typ.r2.num);
	   break;
	 case subinttyp:
	   printf("-%d",prg->u.typ.r2.num);	
	   break;
	 case consttyp:
	   printf("%s",prg->u.typ.r2.id);
	   break;
	 }
       if (prg->next!=NULL)
	 {
	   printf(", ");
	   prtprg(prg->next);
	 }
       switch(prg->u.typ.base)
	 {
	 case intTy:
	   printf("] of int");
	   break;
	 case realTy:
	   printf("] of real");
	   break;
	 case strTy:
	   printf("] of string");
	   break;
	 case boolTy:
	   printf("] of boole");
	   break;
	 }
       break;
     }
 }

 void prtstmt(A_node prg, int mode)
 {
   linebreak(prg);
   switch (prg->u.stmt.kind)
     {
     case subcalls:
       linebreak(prg,1);
       printf("%s(",prg->u.stmt.id);
       prtprg(prg->u.stmt.exprlist);
       if(mode==2)
	 printf(")");
       else
	 printf("); ");
       break;
     case assignstmt:
       linebreak(prg,1);
       prtprg(prg->u.stmt.var);
       printf(" := ");
       prtprg(prg->u.stmt.exprlist);
       printf("; ");
       break;
     case stmtblk:
       printf("\n");
       blocksp();
       printf("begin");
       block++;
       lastnum++;
       linebreak(prg,1);
       prtprg(prg->u.stmt.declist);
       if (prg->u.stmt.declist!=NULL) printf("");
       prtprg(prg->u.stmt.stmtlist,1);
       block--;
       if (mode==0)
	 {
	   printf("\n");
	   blocksp();
	   printf("end");
	   lastnum++;
	 }
       else 
	 {
	   printf("\n");
	   blocksp();
	   printf("end; ");
	   lastnum++;
	 }
       break;
     case readstmt:
       linebreak(prg,1);
       printf("read ");
       prtprg(prg->u.stmt.var);
       printf("; ");
       break;
     case writestmt:
       linebreak(prg,1);
       printf("write ");
       prtprg(prg->u.stmt.exprlist);
       printf("; ");
       break;
     case ifstmt:
       linebreak(prg,1);
       printf("if ");
       block++;
       prtprg(prg->u.stmt.exprlist);
       block--;
       printf(" then ");
       block++;
       if (prg->u.stmt.stmtlist!=NULL)
	 prtprg(prg->u.stmt.stmtlist,mode);
       if (prg->u.stmt.elselist!=NULL)
	 {
	   block--;
	   printf("\n");
	   blocksp();
	   printf("else");
	   block++;
	   lastnum++;
	   prtprg(prg->u.stmt.elselist,mode);
	 }
       block--;
       printf("\n");
       blocksp();
       printf("end if; ");
       lastnum++;
       break;
     case whilestmt:
       linebreak(prg,1);
       printf("while ");
       block++;
       prtprg(prg->u.stmt.exprlist);
       block--;
       printf(" do");
       block++;
       prtprg(prg->u.stmt.stmtlist,mode);
       block--;
       printf("\n");
       blocksp();      
       printf("end while; ");
       lastnum++;
       break;
     case returnstmt:
       linebreak(prg,1);
       printf("return ");
       block++;
       prtprg(prg->u.stmt.exprlist);
       block--;
       printf("; ");
       break;
     }
   if (prg->next!=NULL)
     prtprg(prg->next,mode);
 }

 void prtexp(A_node prg, int mode)
 {
   linebreak(prg,1);
   switch (prg->u.expr.kind)
     {
     case add:
       prtprg(prg->u.expr.operand1);
       printf("+");
       prtprg(prg->u.expr.operand2);
       break;
     case sub:
       prtprg(prg->u.expr.operand1);
       printf("-");
       prtprg(prg->u.expr.operand2);
       break;
     case mult:
       prtprg(prg->u.expr.operand1);
       printf(" * ");
       prtprg(prg->u.expr.operand2);
       break;
     case divide:
       prtprg(prg->u.expr.operand1);
       printf(" / ");
       prtprg(prg->u.expr.operand2);
       break;
     case mod:
       prtprg(prg->u.expr.operand1);
       printf(" %% ");
       prtprg(prg->u.expr.operand2);
       break;
     case lt:
       prtprg(prg->u.expr.operand1);
       printf("<");
       prtprg(prg->u.expr.operand2);
       break;
     case gt:
       prtprg(prg->u.expr.operand1);
       printf(">");
       prtprg(prg->u.expr.operand2);
       break;
     case eq:
       prtprg(prg->u.expr.operand1);
       printf("=");
       prtprg(prg->u.expr.operand2);
       break;
     case not:
       printf("!");
       prtprg(prg->u.expr.operand2);
       break;
     case neq:
       prtprg(prg->u.expr.operand1);
       printf("<>");
       prtprg(prg->u.expr.operand2);
       break;
     case lteq:
       prtprg(prg->u.expr.operand1);
       printf("<=");
       prtprg(prg->u.expr.operand2);
       break;
     case gteq:
       prtprg(prg->u.expr.operand1);
       printf(">=");
       prtprg(prg->u.expr.operand2);
       break;
     case or:
       prtprg(prg->u.expr.operand1);
       printf(" || ");
       prtprg(prg->u.expr.operand2);
       break;
     case and:
       prtprg(prg->u.expr.operand1);
       printf(" && ");
       prtprg(prg->u.expr.operand2);
       break;
     case ex:
       prtprg(prg->u.expr.operand1);
       break;
     }
   if (prg->next!=NULL)
     {
       printf(", ");
       prtprg(prg->next);
     }
 }

 void prtoper(A_node prg, int mode)
 {
   linebreak(prg,1);
   char *st;
   switch (prg->u.oper.kind)
     {
     case number:
       printf("%d",prg->u.oper.u.intval);
       break;
     case real:
       printf("%f",prg->u.oper.u.realval);
       break;
     case str:
       quotparse(prg->u.oper.u.strval,&st);
       printf("\"%s\"",st);
       free(st);
       break;
     case variable:
       prtprg(prg->u.oper.u.var);
       break;
     case subcall:
       prtprg(prg->u.oper.u.subcall,2);
       break;
     case expr:
       printf("(");      
       prtprg(prg->u.oper.u.expr);
       printf(")");
       break;
     }
 }

 void prtvar(A_node prg, int mode)
 {
   linebreak(prg,1);
   switch (prg->u.var.kind)
     {
     case arrvar:
       printf("%s[",prg->u.var.id);
       prtidx(prg);
       printf("]");
       break;
     case varvar:
       printf("%s",prg->u.var.id);
       break;
     }
 }

 void prtid(A_node prg, int mode)
 {
   linebreak(prg,1);
   printf("%s",prg->u.idlist.id);
   if (prg->u.idlist.next!=NULL)
     {
       printf(", ");
       prtprg(prg->u.idlist.next);
     }
   return;
 }

 void prtidx(A_node prg)
 {
   linebreak(prg,1);
   prtprg(prg->u.var.idxlist.expr);
   if (prg->u.var.idxlist.next!=NULL)
     {
       printf(", ");
       prtidx(prg->u.var.idxlist.next);
     }
 }

 /**************************************************************
 Abstract Syntax Tree Traversal Functions
 **************************************************************/
int travprg(A_node prg, int mode=0, A_node *list=NULL,int rettyp=0,A_node retnode=NULL,rettype **val=NULL, rettype **val2=NULL, int *dim=NULL, int **range=NULL)
     /*Main Recursive loopback*/
{
  int i=0,i2=0;
  rettype v,*v1;
  A_node idx;
  DRlist *idx2;

  if (prg==NULL) return 0;

  switch (prg->entity)
    {
    case declTag:
      idx=prg;
      while (idx!=NULL)
	{
	  travdecl(idx,mode,list);
	  if (idx->u.decl.kind==subroutineDecl)
	    idx=idx->u.decl.u.sub.next;
	  else
	    idx=idx->u.decl.next;
	}
      break;
    case typTag:
      i=travtyp(prg,mode);
      break;
    case stmtTag:
      idx=prg;
      while (idx!=NULL)
	{
	  if (val!=NULL)
	    i2=travstmt(idx,mode,rettyp,val);
	  else
	    i2=travstmt(idx,mode,rettyp,&v1);
	  if (idx==prg)
	    i=i2;
	  idx=idx->next;
	}
      break;
    case expTag:
      if (retnode!=NULL)
	{
	  idx2=htab.search(retnode->u.stmt.id)->r->args;
	}
      idx=prg;
      while (idx!=NULL)
	{
	  if (retnode==NULL)
	    {
	      if (val!=NULL)
		i2=travexp(idx,mode,val,dim,range);
	      else
		i2=travexp(idx,mode,&v1,dim,range);
	    }
	  else
	    {
	      if (val!=NULL)
		i2=travexp(idx,mode,val,dim,range,idx2->typ);
	      else
		i2=travexp(idx,mode,&v1,dim,range,idx2->typ);
	    }
	  if (prg==idx) i=i2;
	  if ((i2&taINT)==taINT)
	    i=i|taINT;
	  idx=idx->next;
	  if (retnode!=NULL) idx2=idx2->next;
	}
      break;
    case operTag:
      i=travoper(prg,mode,val,dim,range);
      break;
    case varTag:
      i=travvar(prg,mode,val,dim,range);
      break;
    case idTag:
      if (val==NULL)
	travid(prg,mode,list,rettyp,retnode,NULL);
      else
	travid(prg,mode,list,rettyp,retnode,*val);
      break;
    case undefTag:
    default:
      printf("Error on line [%d]! [%d]\n",prg->lineno,prg->entity);
      globerr=-1;
      return -2;
      break;
    }
  return i;
}

void travdecl(A_node prg, int mode, A_node *list)
{
  rettype *v,*val2;
  char *st;
  A_node lst,lst2,dectyp;
  int i1, low, high;

  v=(rettype *)malloc(sizeof(rettype));
  switch (prg->u.decl.kind)
    {
    case progDecl:
      lst=NULL;
      lst2=NULL;
      decl(prg->u.decl.id,prg,&lst2,0x0ffff);
      genprgname(prg->u.decl.id);
      htab.pushScope();     
      travprg(prg->u.decl.u.prog.declist,2|2048|16384,&lst);
      gennewline();
      genstdinit();    
      travprg(prg->u.decl.u.prog.subdecl,0,&lst);
      gennewline();
      genmainhead();
      currloc=1;
      travprg(prg->u.decl.u.prog.stmtblk,0|4096,NULL,0x0ffff);
      gennewline();
      genmainend();
      htab.popScope(lst,1);
      htab.popScope(lst2,0);
      freelist(lst);
      freelist(lst2);
      break;
    case varDecl:
      if ((i1=travprg(prg->u.decl.u.var.typ,1))!=-1)
	{
	  getrange(prg->u.decl.u.var.typ,&low,&high);
	  travprg(prg->u.decl.u.var.idList,mode,list,i1|tVAR,prg->u.decl.u.var.typ);
	}
      break;
    case constDecl:
      switch(prg->u.decl.u.cnst.kind)
	{
	case realtyp:
	  v->kind=realTy;
	  v->lineno=prg->lineno;
	  v->u.realval=prg->u.decl.u.cnst.val.realval;
	  travprg(prg->u.decl.u.cnst.idList,mode,list,tVAR|tCONST|tREAL,NULL,&v);
	  break;
	case inttyp:
	  v->kind=intTy;
	  v->lineno=prg->lineno;
	  v->u.intval=prg->u.decl.u.cnst.val.intval;
	  travprg(prg->u.decl.u.cnst.idList,mode,list,tVAR|tCONST|tINT,NULL,&v);
	  break;
	case strtyp:
	  v->kind=strTy;
	  v->lineno=prg->lineno;
	  v->u.strval=prg->u.decl.u.cnst.val.strval;
	  travprg(prg->u.decl.u.cnst.idList,mode,list,tVAR|tCONST|tSTRING,NULL,&v);
	  break;
	case subrealtyp:
	  v->kind=realTy;
	  v->lineno=prg->lineno;
	  v->u.realval=-prg->u.decl.u.cnst.val.realval;
	  travprg(prg->u.decl.u.cnst.idList,mode,list,tVAR|tCONST|tREAL,NULL,&v);
	  break;
	case subinttyp:
	  v->kind=intTy;
	  v->lineno=prg->lineno;
	  v->u.intval=-prg->u.decl.u.cnst.val.intval;
	  travprg(prg->u.decl.u.cnst.idList,mode,list,tVAR|tCONST|tINT,NULL,&v);
	  break;
	case addrealtyp:
	  v->kind=realTy;
	  v->lineno=prg->lineno;
	  v->u.realval=prg->u.decl.u.cnst.val.realval;
	  travprg(prg->u.decl.u.cnst.idList,mode,list,tVAR|tCONST|tREAL,NULL,&v);
	  break;
	case addinttyp:
	  v->kind=intTy;
	  v->lineno=prg->lineno;
	  v->u.intval=prg->u.decl.u.cnst.val.intval;
	  travprg(prg->u.decl.u.cnst.idList,mode,list,tVAR|tCONST|tINT,NULL,&v);
	  break;
	case consttyp:
	  if ((i1=varaccess(prg->u.decl.id,&dectyp,&val2, prg->lineno))!=-1)
	    {
	      if ((i1&tCONST)==tCONST)
		{
		  val2->lineno=prg->lineno;
		  travprg(prg->u.decl.u.cnst.idList,mode,list,i1,dectyp,&val2);
		}
	      else
		{
		  printf("Error[%d], cannot set a constant to a variable type\n", prg->lineno);
		  globerr=-1;
		}
	    }
	  break;
	default:
	  break;
	}
      break;
    case subroutineDecl:
      lst=NULL;
      i1=0;
      if (prg->u.decl.u.sub.rettype!=NULL)
	{      
	  if (prg->u.decl.u.sub.rettype->u.typ.base!=0)
	    {
	      i1=travprg(prg->u.decl.u.sub.rettype);
	    }
	}
      else
	i1=0x0ff;
      i1=i1|tFUNCTION;
      decl(prg->u.decl.id,prg->u.decl.u.sub.rettype,&lst,i1,NULL,prg->lineno);
      gensubhead1(prg->u.decl.id);
      subcomplete(prg->u.decl.id,prg->u.decl.u.sub.vallist,prg->u.decl.u.sub.reflist);
      gensubhead2(i1);
      htab.pushScope();
      currloc=0;
      travprg(prg->u.decl.u.sub.vallist,0,&lst);
      if (prg->u.decl.u.sub.reflist!=NULL)
	{      
	  travprg(prg->u.decl.u.sub.reflist,0,&lst);
	}
      /*      prtlist(lst);*/
      travprg(prg->u.decl.u.sub.stmtblk,0|4096,NULL,i1);
      if(i1&0xf0==taINT)
	{
	  printf("Error[%d], cannot return ARRAYS.\n",prg->lineno);
	  globerr=-1;
	}
      htab.popScope(lst,1);
      freelist(lst);
      gensubend(i1);
      break;
    }
}

int travtyp(A_node prg, int mode)
{
  int i=0;
  switch(prg->u.typ.kind)
    {
    case baseTy:
      switch(prg->u.typ.base)
	{
	case intTy:
	  i=(i&3)|tINT;
	  break;
	case realTy:
	  i=(i&3)|tREAL;
	  break;
	case strTy:
	  i=(i&3)|tSTRING;
	  break;
	case boolTy:
	  i=(i&3)|tBOOLE;
	  break;
	}
      break;
    case array1Ty:
      switch(prg->u.typ.base)
	{
	case intTy:
	  i=(i&3)|taINT;
	  break;
	case realTy:
	  i=(i&3)|taREAL;
	  break;
	case strTy:
	  i=(i&3)|taSTRING;
	  break;
	case boolTy:
	  i=(i&3)|taBOOLE;
	  break;
	default:
	  printf("Error[%d], invalid base type for array.\n",prg->lineno);
	  globerr=-1;
	  break;
	}
      switch(prg->u.typ.base1)
	{
	case consttyp:
	  i=(i&0x0ffffc)|varaccess(prg->u.typ.r1.id,NULL, NULL, prg->lineno);
	  break;
	}
      if ((prg->next!=NULL) && (i&3==0))
	{
	  i=(i&0x0ffffc)|travprg(prg->next);
	}
      break;
    case array2Ty:
      switch(prg->u.typ.base)
	{
	case intTy:
	  i=(i&3)|taINT;
	  break;
	case realTy:
	  i=(i&3)|taREAL;
	  break;
	case strTy:
	  i=(i&3)|taSTRING;
	  break;
	case boolTy:
	  i=(i&3)|taBOOLE;
	  break;
	default:
	  printf("Error[%d], invalid base type for array.\n",prg->lineno);
	  globerr=-1;
	  break;
	}
      switch(prg->u.typ.base1)
	{
	case consttyp:
	  i=(i&0x0ffffc)|varaccess(prg->u.typ.r1.id,0,0, prg->lineno);
	  break;
	}
      if (i&3==1)
	return i;
      switch(prg->u.typ.base2)
	{
	case consttyp:
	  i=(i&0x0ffffc)|varaccess(prg->u.typ.r2.id,0,0, prg->lineno);
	  break;
	}
      if ((prg->next!=NULL) && (i==0))
	{
	  i=(i&0x0ffffc)|travprg(prg->next);
	}
      break;
    }
  return i;
}

int travstmt(A_node prg, int mode, int rettyp, rettype **val)
{
  int i=0,i1,i2;
  int err;
  int dim;
  int *range=NULL;
  char lbl0[6];
  char lbl1[6];
  char lbl2[6];
  A_node lst;  

  switch (prg->u.stmt.kind)
    {
    case subcalls:
      i=varaccess(prg->u.stmt.id,NULL,NULL,prg->lineno);
      if ((i==0x0ffff)||((i&tFUNCTION)!=tFUNCTION))
	{
	  printf("Error[%d], %s is not a valid function\n",prg->lineno,prg->u.stmt.id);
	  globerr=-1;
	}
      else
	{
	  travprg(prg->u.stmt.exprlist,1|8192,0,0,prg);
	  checksubcall(prg->u.stmt.id,prg->u.stmt.exprlist);
	  i=i|64;
	  if (val!=NULL)
	    {
	      (*val)=(rettype *)malloc(sizeof(rettype));
	      (*val)->id=prg->u.stmt.id;
	    }
	}
      break;
    case assignstmt:
      i2=travprg(prg->u.stmt.exprlist,1|8192,0,0,0,0,0,&dim,&range);
      i1=travprg(prg->u.stmt.var,1);
      gencast(i1,i2);
      i1=travprg(prg->u.stmt.var,1|256);
      if ((i1&tVAR)!=tVAR)
	{
	  printf("Error[%d], cannot assign values to non variables\n",prg->lineno);
	  globerr=-1;
	}
      err=0;
      if ((i1&taBOOLE)!=(i2&taBOOLE))
	{
	  switch(i2&taBOOLE)
	    {
	    case tBOOLE:
	      err=-1;
	      break;
	    case tSTRING:
	      switch(i1&taBOOLE)
		{
		case tINT:
		case tREAL:
		case taINT:
		case taREAL:
		case taSTRING:
		case taBOOLE:
		  err=-1;
		  break;
		}
	      break;
	    case tREAL:
	      switch (i1&taBOOLE)
		{
		case tSTRING:
		case taSTRING:
		case taBOOLE:
		case taINT:
		case taREAL:
		  err=-1;
		  break;
		}
	      break;
	    case tINT:
	      switch (i1&taBOOLE)
		{
		case tSTRING:
		case taSTRING:
		case taBOOLE:
		case taINT:
		case taREAL:
		  err=-1;
		  break;
		}
	      break;
	    case taINT:
	      switch (i1&taBOOLE)
		{
		case taSTRING:
		case tSTRING:
		case tBOOLE:
		case tINT:
		case tREAL:
		  err=-1;
		  break;
		}
	      break;
	    case taREAL:
	      switch (i1&taBOOLE)
		{
		case taSTRING:
		case tSTRING:
		case tBOOLE:
		case tINT:
		case tREAL:
		  err=-1;
		  break;
		}
	      break;
	    case taSTRING:
	      switch (i1&taBOOLE)
		{
		case taINT:
		case taREAL:
		case tSTRING:
		case tBOOLE:
		case tINT:
		case tREAL:
		  err=-1;
		  break;
		}
	      break;
	    case taBOOLE:
	      err=-1;
	      break;
	    }
	}
      if (((i1&taINT)==taINT)&&((i2&taINT)==taINT))
	{
	  err=dimmatch(prg->u.stmt.var,prg->u.stmt.exprlist,&dim,&range, prg->u.stmt.var->u.var.id,"Expression");
	}
      if (err==-1)
	{
	  printf("Error[%d], cannot cast a [",prg->lineno);
	  prtexpand(i2);
	  printf("] into a type [");
	  prtexpand(i1);
	  printf("]\n");
	  globerr=-1;
	}
      break;
    case stmtblk:
      htab.pushScope();
      lst=NULL;
      travprg(prg->u.stmt.declist,mode,&lst);
      travprg(prg->u.stmt.stmtlist,0,NULL,rettyp);
      htab.popScope(lst,1);
      freelist(lst);
      break;
    case readstmt:
      i1=travprg(prg->u.stmt.var,1|512|256);
      if((i1&taINT)==taINT)
	{
	  printf("Error[%d], cannot read whole arrays\n",prg->lineno);
	  globerr=-1;
	}
      break;
    case writestmt:
      i1=travprg(prg->u.stmt.exprlist,1|1024|8192);
      if((i1&taINT)==taINT)
	{
	  printf("Error[%d], cannot write whole arrays\n",prg->lineno);
	  globerr=-1;
	}
      break;
    case ifstmt:
      i1=travprg(prg->u.stmt.exprlist,1|8192);
      if((i1&taINT)==taINT)
	{
	  printf("Error[%d], conditions cannot be arrays\n",prg->lineno);
	  globerr=-1;
	}
      genbcast();
      genifstmt((char *)lbl0);
      if (prg->u.stmt.stmtlist!=NULL)
	travprg(prg->u.stmt.stmtlist,0,NULL,rettyp);
      genelsestmt(lbl0,(char *)lbl1);
      if (prg->u.stmt.elselist!=NULL)
	travprg(prg->u.stmt.elselist,0,NULL,rettyp);
      genendif(lbl1);
      break;
    case whilestmt:
      genwhilebgn((char *)lbl0);
      i1=travprg(prg->u.stmt.exprlist,1|8192);
      if((i1&taINT)==taINT)
	{
	  printf("Error[%d], conditions cannot be arrays\n",prg->lineno);
	  globerr=-1;
	}
      genbcast();
      /*casting*/
      genwhilechk((char *)lbl1);
      travprg(prg->u.stmt.stmtlist,mode,NULL,rettyp);
      genwhileend(lbl0,lbl1);
      break;
    case returnstmt:
      i1=travprg(prg->u.stmt.exprlist,1|8192);
      err=0;
      if ((rettyp==0x0ffff)&&(prg->u.stmt.exprlist!=NULL))
	{
	  printf("Error[%d], program cannot return a value\n",prg->lineno);
	  err=-1;
	  globerr=-1;
	}
      if ((rettyp==0x0ff)&&(prg->u.stmt.exprlist!=NULL))
	{
	  printf("Error[%d], cannot return a value for this subroutine\n",prg->lineno);
	  err=-1;
	  globerr=-1;
	}
      if (( (rettyp!=0x0ffff)&&(rettyp!=0x0ff) )&&(prg->u.stmt.exprlist==NULL))
	{
	  printf("Error[%d], subroutine requires a return value\n",prg->lineno);
	  err=-1;
	  globerr=-1;
	}
      err=cast(rettyp,i1,err,prg->lineno);
      gencast(rettyp,i1);
      if ((i1&0xf0)==taINT)
	{
	  printf("Error[%d], cannot return ARRAYS.\n",prg->lineno);
	  i=-1;
	  globerr=-1;
	}
      break;
    }
  return i;
}

int travexp(A_node prg, int mode, rettype **rval, int *dim=NULL, int **range=NULL, int paramtyp=-1)
{
  /*handle all expression type checking*/
  int typ1=0,typ2=0,result=0;
  int dim1,dim2,*range1=NULL,*range2=NULL;
  rettype *val2=NULL, *val1=NULL;
  A_node getter;
  char *iden1,*iden2;

  typ2=travprg(prg->u.expr.operand2,mode&3);
  switch (prg->u.expr.kind)
    {
    case add:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3)
	{
	  genop1cast(typ1,typ2);
	}
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3)
	{
	  genop2cast(typ1,typ2);
	}
      result=chkadd(typ1,typ2, prg->lineno,val1,val2,rval);
      if (mode>3)
	{
	  genaddcode(result);
	}
      break;
    case sub:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3)
	{
	  genop1cast(typ1,typ2);
	}
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3)
	{
	  genop2cast(typ1,typ2);
	}
      result=chksub(typ1,typ2, prg->lineno,val1,val2,rval);
      if (prg->u.expr.operand1!=NULL)
	{
	  if (mode>3) gensubcode(result, 2);
	}
      else
	{
	  if (mode>3) gensubcode(result, 1);
	}
      break;
    case mult:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3)
	{
	  genop1cast(typ1,typ2);
	}
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3)
	{
	  genop2cast(typ1,typ2);
	}
      result=chkmult(typ1,typ2, prg->lineno,val1,val2,rval);
      if (mode>3) genmultcode(result);
      break;
    case divide:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3)
	{
	  genop1cast(typ1,typ2);
	}
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3)
	{
	  genop2cast(typ1,typ2);
	}
      result=chkdiv(typ1,typ2, prg->lineno,val1,val2,rval);
      if (mode>3) gendivcode(result);
      break;
    case mod:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3) gencast(tINT,typ1);
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3) gencast(tINT,typ1);
      result=chkmod(typ1,typ2, prg->lineno,val1,val2,rval);
      if (mode>3) genmodcode(result);
      break;
    case lt:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3)
	{
	  genop1cast(typ1,typ2);
	}
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3)
	{
	  genop2cast(typ1,typ2);
	}
      result=chkrel(typ1,typ2, prg->lineno,"<",val1,val2,rval);
      if (((typ1&taBOOLE)==tREAL)||((typ2&taBOOLE)==tREAL))
	{
	  if (mode>3) genltcode(tREAL);
	}
      else
	{
	  if (mode>3) genltcode(result);
	}
      break;
    case gt:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3)
	{
	  genop1cast(typ1,typ2);
	}
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3)
	{
	  genop2cast(typ1,typ2);
	}
      result=chkrel(typ1,typ2, prg->lineno,">",val1,val2,rval);
      if (((typ1&taBOOLE)==tREAL)||((typ2&taBOOLE)==tREAL))
	{
	  if (mode>3) gengtcode(tREAL);
	}
      else
	{
	  if (mode>3) gengtcode(result);
	}
      break;
    case eq:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3)
	{
	  genop1cast(typ1,typ2);
	}
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3)
	{
	  genop2cast(typ1,typ2);
	}
      result=chkrel(typ1,typ2, prg->lineno,"=",val1,val2,rval);
      if (mode>3) geneqcode(result);
      break;
    case not:
      typ1=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3) genbcast(typ1);
      result=chknot(typ1, prg->lineno,val1,rval);
      if (mode>3) gennotcode(result);
      break;
    case neq:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3)
	{
	  genop1cast(typ1,typ2);
	}
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3)
	{
	  genop2cast(typ1,typ2);
	}
      result=chkrel(typ1,typ2, prg->lineno,"<>",val1,val2,rval);
      if (mode>3) genneqcode(result);
      break;
    case lteq:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3)
	{
	  genop1cast(typ1,typ2);
	}
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3)
	{
	  genop2cast(typ1,typ2);
	}
      result=chkrel(typ1,typ2, prg->lineno,"<=",val1,val2,rval);
      if (((typ1&taBOOLE)==tREAL)||((typ2&taBOOLE)==tREAL))
	{
	  if (mode>3) genlteqcode(tREAL);
	}
      else
	{
	  if (mode>3) genlteqcode(result);
	}
      break;
    case gteq:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3)
	{
	  genop1cast(typ1,typ2);
	}
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3)
	{
	  genop2cast(typ1,typ2);
	}
      result=chkrel(typ1,typ2, prg->lineno,">=",val1,val2,rval);
      if (((typ1&taBOOLE)==tREAL)||((typ2&taBOOLE)==tREAL))
	{
	  if (mode>3) gengteqcode(tREAL);
	}
      else
	{
	  if (mode>3) gengteqcode(result);
	}
      break;
    case or:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3) genbcast(typ1);
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3) genbcast(typ2);
      result=chklog(typ1,typ2, prg->lineno,"||",val1,val2,rval);
      if (mode>3) genorcode(result);
      break;
    case and:
      typ1=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,&val1,0,&dim1,&range1);
      if (mode>3)
	{
	  genbcast(typ1);
	}
      typ2=travprg(prg->u.expr.operand2,(mode)&0x0FAFF,0,0,0,&val2,0,&dim2,&range2);
      if (mode>3)
	{
	  genbcast(typ2);
	}
      result=chklog(typ1,typ2, prg->lineno,"&&",val1,val2,rval);
      if (mode>3)
	{
	  genandcode(result);
	}
      break;
    case ex:
      result=travprg(prg->u.expr.operand1,(mode)&0x0FAFF,0,0,0,rval,0,&dim1,&range1);
      typ1=result;
      break;
    }
  if ((typ1!=0x0ff)&&(typ2!=0x0ff))
    {
      if(((typ2&taINT)==taINT)&&((typ1&taINT)==taINT))
	{
	  getter=prg->u.expr.operand1->u.oper.u.var;
	  iden1=getter->u.var.id;
	  getter=prg->u.expr.operand2->u.oper.u.var;
	  iden2=getter->u.var.id;
	  drmatch(dim1,range1,dim2,range2,prg->lineno,iden1,iden2);
	}
      if((dim!=NULL) && (range!=NULL))
	{
	  if((typ2&taINT)==taINT)
	    {
	      *dim=dim2;
	      *range=range2;
	    }
	  if((typ1&taINT)==taINT)
	    {
	      *dim=dim1;
	      *range=range1;
	    }
	}
    }
  if ((mode&1024)==1024)
    {
      if (prg->next==NULL)
	genwritelnstmt(result);
      else
	genwritestmt(result);
    }
  if (paramtyp!=-1)
    {
      gencast(paramtyp,result);
    }
  return result;
}

int travoper(A_node prg, int mode, rettype **val,int *dim,int **range)
{
  int i=0;
  int location;
  
  switch (prg->u.oper.kind)
    {
    case number:
      i=0|tCONST;
      if (val!=NULL)
	{
	  *val=(rettype *)malloc(sizeof(rettype));
	  (*val)->kind=intTy;
	  (*val)->u.intval=prg->u.oper.u.intval;
	  if((mode&8192)==8192) genconsti((*val)->u.intval);
	}
      break;
    case real:
      i=4|tCONST;
      if (val!=NULL)
	{
	  *val=(rettype *)malloc(sizeof(rettype));
	  (*val)->kind=realTy;
	  (*val)->u.realval=prg->u.oper.u.realval;
	  if ((mode&8192)==8192) genconstf(val);
	}
      break;
    case str:
      if (val!=NULL)
	{
	  *val=(rettype *)malloc(sizeof(rettype));
	  (*val)->kind=strTy;
	  (*val)->u.strval=prg->u.oper.u.strval;
	  if ((mode&8192)==8192) genconsts((*val)->u.strval);
	}
      i=8|tCONST;
      break;
    case variable:
      i=travprg(prg->u.oper.u.var,mode,0,0,0,val,0,dim,range);
      if (i!=-1)
      {
	if((i&0x0fffff)==0x0ffff)
	  {
	    i=-1;
	    globerr=-1;
	    printf("Error[%d], cannot call a program\n",prg->lineno);
	  }      
	else if((i&tFUNCTION)==tFUNCTION)
	  {
	    i=-1;
	    globerr=-1;
	    printf("Error[%d], cannot use function as variable\n",prg->lineno);
	  }
      }
      if ((val!=NULL)&&((*val)==NULL))
	{
	  (*val)=(rettype *)malloc(sizeof(rettype));
	  if((i&(tCONST|tVAR))==tCONST)
	    {
	      switch(i&taBOOLE)
		{
		case tINT:
		  if ((mode&8192)==8192) genconsti((*val)->u.intval);
		  break;
		case tREAL:
		  if ((mode&8192)==8192) genconstf(val);
		  break;
		case tSTRING:
		  if ((mode&8192)==8192) genconsts((*val)->u.strval);
		  break;
		}
	    }
	  else if(((i&tFUNCTION)!=tFUNCTION)&&((i&taINT)!=taINT))
	    {
	    }
	  (*val)->id=(char *) malloc(strlen(prg->u.oper.u.var->u.var.id)+1);
	  strcpy((*val)->id,prg->u.oper.u.var->u.var.id);
	}
      break;
    case subcall:
      i=travprg(prg->u.oper.u.subcall,mode,0,0,0,val);
      if (i!=-1)
	{
	  if ((i&0x0fffff)==0x0ff)
	    {
	      printf("Error[%d], subroutine does not have a value\n",prg->lineno);
	      globerr=-1;
	      i=-1;
	    }
	  if ((i&0x0fffff)==0x0ffff)
	    {
	      printf("Error[%d], program cannot return a value\n",prg->lineno);
	      i=-1;
	      globerr=-1;
	    }
	}
      break;
    case expr:
      i=travprg(prg->u.oper.u.expr,mode|8192,0,0,0,val,0,dim,range);
      break;
      }
  return i;
}
 
int travvar(A_node prg, int mode, rettype **val,int *dim, int **range)
{
  int i=0;
  A_node typ;
  int loc;
  switch (prg->u.var.kind)
    {
    case arrvar:
      if (mode>=1) i=varaccess(prg->u.var.id,&typ,0, prg->lineno,0,0,&loc);
      if ((i&taINT)!=taINT)
	{
	  printf("Error[%d], %s is not an array\n",prg->lineno,prg->u.var.id);
	  i=-1;
	  globerr=-1;
	}
      else
	{
	  chkarry(prg,typ);
	  i=(i&0x0ffffef);
	  if ((mode&512)==512)
	    {
	      genreadstmt(i);
	    }
	  if ((mode&256)==256) genarrstore1(loc);
	  if ((mode&8192)==8192) genarrload1(loc);	  
	  travidx(prg,mode,0,prg->u.var.id);
	  if ((mode&256)==256) genarrstore2(i);
	  if ((mode&8192)==8192) genarrload2(i);
	}
      break;
    case varvar:
      if (mode>=1)
	{
	  i=varaccess(prg->u.var.id,&typ,val, prg->lineno,dim,range,&loc);
	  if(loc==-1)
	    {
	      if ((mode&8192)==8192) genglobacc(symtable->namestr,prg->u.var.id,i);
	    }
	  else
	    {
	      if ((mode&8192)==8192) genlocacc(i,loc);
	    }
	}
      if ((mode&512)==512)
	{
	  genreadstmt(i);
	}
      if ((mode&256)==256) genstore(i,loc,prg->u.var.id,symtable->namestr);
      break;
    }
  return i;
}

void travid(A_node prg, int mode, A_node *list,int typ, A_node retnode,rettype *val)
{
  int dim, *range, loc;
  int i, high,low;
  rettype v;

  if ((mode&1)==1)
    {
      varaccess(prg->u.idlist.id,0,0, prg->lineno);
    }
  else
    {
      if((typ&taINT)==taINT)
	{
	  dim=getdim(retnode);
	  range=(int *)malloc(dim*sizeof(int));
	  for(i=0;i<dim;i++)
	    {
	      getrange(retnode,&low,&high,i,retnode->lineno);
	      range[i]=high-low+1;
	      if ((mode&4096)==4096) genconsti(range[i]);
	    }
	   if ((mode&4096)==4096) gennewarr(typ,dim, currloc);
	}
      else
	{
	  dim=0;
	  range=NULL;
	}
      if (val!=NULL)
	{
	  if ((mode&2048)==2048) genglob(prg->u.idlist.id,typ,*val,dim,range);
	  if ((mode&16384)==16384) genbgninc(symtable->namestr, prg->u.idlist.id,typ,*val,dim,range);
	}
      else
	{
	  if ((mode&2048)==2048) genglob(prg->u.idlist.id,typ,v,dim,range);
	  if ((mode&16384)==16384) genbgninc(symtable->namestr, prg->u.idlist.id,typ,v,dim,range);
	}
      if ((mode&8192)==8192)
	{
	  decl(prg->u.idlist.id,retnode,list,typ,val);
	}
      else if ((typ&tVAR)==tVAR)
	{
	  decl(prg->u.idlist.id,retnode,list,typ,val,0,currloc);
	  currloc++;
	}
      else
	{
	  decl(prg->u.idlist.id,retnode,list,typ,val);
	}
    }
  if (prg->u.idlist.next!=NULL)
    {
      travprg(prg->u.idlist.next,mode,list,typ,retnode);
    }
  return;
}

int getmin(char *id, int dim)
{
  A_node typ;
  int low,high;
  varaccess(id,&typ);
  getrange(typ,&low,&high,dim);
  return low;
}

void travidx(A_node prg, int mode, int dim=0, char *id=NULL)
{
  int min;
  /*printf("%d\n",prg->u.var.idxlist.kind);*/
  if ((mode&(256|8192))!=0)
    {
      travprg(prg->u.var.idxlist.expr,1|8192);
      min=getmin(id, dim);
      genconsti(min);
      gensubcode(tINT,2);
    }
  if (prg->u.var.idxlist.next!=NULL)
    {
      genarridx(mode,0);
      travidx(prg->u.var.idxlist.next, mode, dim+1, id);
    }
  else
    genarridx(mode,1);

}

void subcomplete(char *id,A_node val, A_node ref)
{
  A_node idx,idx2;
  int dim,high,low,typ;
  int *range;
  int cnt;

  idx = val;

  while (idx!=NULL)
    {
      switch(idx->u.decl.kind)
	{
	case varDecl:
	  dim=getdim(idx->u.decl.u.var.typ);
	  range=(int *)malloc(dim);
	  for(cnt=0;cnt<dim;cnt++)
	    {
	      getrange(idx->u.decl.u.var.typ,&low,&high,cnt);
	      range[cnt]=high-low+1;
	    }	  
	  typ=travprg(idx->u.decl.u.var.typ);
	  idx2=idx->u.decl.u.var.idList;
	  while (idx2!=NULL)
	    {
	      gensubheadmid(typ,dim);
	      update(id,idx2->u.idlist.id,0,typ,dim,&range);
	      idx2=idx2->u.idlist.next;
	    }
	  break;
	default:
	  break;
	}
      idx=idx->u.decl.next;
    }
  idx=ref;
  while (idx!=NULL)
    {
      switch(idx->u.decl.kind)
	{
	case varDecl:
	  dim=getdim(idx->u.decl.u.var.typ);
	  range=(int *)malloc(dim);
	  for(cnt=0;cnt<dim;cnt++)
	    {
	      getrange(idx->u.decl.u.var.typ,&low,&high,cnt);
	      range[cnt]=high-low+1;
	    }
	  typ=travprg(idx->u.decl.u.var.typ);
	  if ((typ&taINT)==taINT)
	    {
	      printf("Error[%d], cannot pass array by reference\n",idx->lineno);
	      globerr=-1;
	    }
	  idx2=idx->u.decl.u.var.idList;
	  while (idx2!=NULL)
	    {
	      update(id,idx2->u.idlist.id,1,typ,dim,&range);
	      idx2=idx2->u.idlist.next;
	    }
	  break;
	default:
	  break;
	}
      idx=idx->u.decl.next;
    }
}

#endif
