/****************************************************************************
include.h Main structure declaration header

Declares A_node as a pointer to A_Node
Declares A_Node as structure
Declares SYMTAB_REC as structure
Declares SYMTAB_ENTRY as structure of SYMTAB_REC
Declares rettype as structure for yylval return type
Declares HashTab class for hashing
****************************************************************************/

#ifndef include_h
#define include_h

#include <search.h>
#include <iostream.h>

#define symb_dump_off
#define source_list_off

/*****************************************************************************
Global Variables
*****************************************************************************/
int globerr=0;
FILE *fout;
int currloc;
char currlabel[6]="Lb000";
char *srcbgn=NULL;

/*****************************************************************************
Abstract Syntax Tree Structure and Enumerations
*****************************************************************************/
typedef enum { undefTag = 500, declTag, typTag, stmtTag, expTag, operTag,
	       varTag, idTag } tag;
typedef enum { intTy=1000, realTy, strTy, boolTy } ty;
typedef enum { baseTy=1500,array1Ty, array2Ty } varty;
typedef enum { progDecl = 2000, varDecl, constDecl, subroutineDecl } declty;
typedef enum { realtyp=2500, inttyp, strtyp, subrealtyp, subinttyp, addrealtyp,
	       addinttyp, consttyp } basety;
typedef enum { subcalls = 3000, assignstmt, stmtblk, readstmt, writestmt,
	       ifstmt, whilestmt, returnstmt } stmtty;
typedef enum { add = 3500, sub, mult, divide, mod, lt, gt, eq, not, neq,
	       lteq, gteq, or, and, ex} operty;
typedef enum { number = 4000, real, str, variable, subcall, expr } valty;
typedef enum { var =4500, num, subnum } varnumty;
typedef enum { arrvar = 5000, varvar } arrvarty;

typedef struct A_Node *A_node;

/****************************************************************************
Bit Field Value names
****************************************************************************/
#define tINT          0
#define tREAL         4
#define tSTRING       8
#define tBOOLE        12
#define taINT         16
#define taREAL        20
#define taSTRING      24
#define taBOOLE       28
#define tCONST        32
#define tVAR          64
#define tFUNCTION     128
/***************** A_Node ********************/
struct A_Node
{
  tag entity;
  int lineno;
  
  union {
    struct {
      varty kind;
      ty base;
      basety base1, base2;
      union {
	char * id;
	int num;
      } r1;
      union {
	char * id;
	int num;
      } r2;
    } typ;

    struct {
      declty kind;
      char *id;
      union {
	struct { A_node idList, typ; } var;
	struct {
	  basety kind;
	  union{double realval; int intval; char *strval; } val;  
	  A_node idList;
	} cnst;
	struct{
	  A_node declist, subdecl, stmtblk;
	} prog;
	struct{
	  A_node rettype;
	  A_node reflist,vallist,stmtblk;
	  A_node next;
	}sub;
      } u;
      A_node next;
    } decl;

    struct {
      stmtty kind;
      char * id;
      A_node exprlist, var, declist, stmtlist, elselist;
    } stmt;

    struct {
      operty kind;
      A_node operand1, operand2, typ;
    } expr;

    struct {
      valty kind;
      A_node typ;
      union {
	int intval;
	float realval;
	char * strval;
	A_node var, subcall, expr;
      } u;
    } oper;

    struct {
      arrvarty kind;
      char * id;
      struct {
	A_node expr;
	A_node next;
      } idxlist;
      A_node typ;
    } var;

    struct {
      char *id;
      A_node next;
    } idlist;
  } u;
  A_node next;
  friend void A_prtProg(A_node);
};


/**************************************************************
Symbol Table Structures
**************************************************************/
int levelno=0;

typedef struct SymTabEntry SYMTAB_ENTRY;

  typedef struct
  {
    int lineno;
    ty kind;
    char *id;
    union
    {
      double realval;
      int intval;
      char *strval;
      SYMTAB_ENTRY *sym;
      A_node anode;
    }u;
  }rettype;

  typedef struct drlist
  {
    char *id;
    int typ;
    int dim;
    int *range;
    struct drlist *next;
  }DRlist;

  typedef struct SymTabRec
  {
    int levelno;
    int typ;
    int dim;
    int *range;
    DRlist *args;
    int sloc;
    A_node decl;
    struct SymTabRec *prev;
    struct SymTabEntry *top;
    rettype *value;
  }SYMTAB_REC;
    

  typedef struct SymTabEntry
  {
    char *namestr;
    SYMTAB_REC *r;
    SymTabEntry();
    SymTabEntry(char *, SYMTAB_REC *);
    SYMTAB_REC *get();
    int put(int);
  }SYMTAB_ENTRY;


const int MAX_ENTRIES = 1024;
SYMTAB_ENTRY symtable[MAX_ENTRIES];
SYMTAB_ENTRY *symtabPtr = symtable;

/***************************************************
Hash Table class
***************************************************/

class HashTab
{
 public:
  HashTab();
  ~HashTab();

  SYMTAB_ENTRY *search(char *);
  void dump_symTab();
  void pushScope();
  void popScope(A_Node *, int);
};

/**************************************************
Symbol Table member function Definitions
**************************************************/

SymTabEntry::SymTabEntry(){namestr=NULL; r=NULL;}
SymTabEntry::SymTabEntry(char *s, SYMTAB_REC *rr) { namestr = s; r=rr;}

SYMTAB_REC *
SymTabEntry::get()
{
  return (SYMTAB_REC *)r;
}

int SymTabEntry::put(int lev)
{
  ENTRY item;
  item.key = namestr;
  item.data=(char *) this;
  (void)hsearch(item, ENTER);
  return 1;
}

/*************************************************
Hash Table member function Definitions
*************************************************/
HashTab::HashTab() {(void) hcreate(MAX_ENTRIES);}

HashTab::~HashTab() {hdestroy();}

SYMTAB_ENTRY * HashTab::search(char *namestr)
{
  ENTRY item;
  item.key=namestr;

  ENTRY *found;
  if ((found=hsearch(item,FIND))!=NULL)
    return (SYMTAB_ENTRY *)(found->data);
  else return NULL;
}

void HashTab::dump_symTab()
{
  cout << "SymbolTable snapshot. Current level number: " << levelno << endl;

  for (register SYMTAB_ENTRY *e = symtable, *n=symtabPtr; e!=n; e++)
    if (e->r !=NULL)
      {
	cout << "\tNamestr: " << e->namestr << " levelno:";
	for (SYMTAB_REC *r = e->r ; r; r=r->prev)
	  cout << " " << r->levelno;
	cout << endl;
      }
}

void HashTab::pushScope() {levelno++;}

void HashTab:: popScope(A_node decls, int lev)
{
  A_node idx;
  SYMTAB_ENTRY *found;
  SYMTAB_REC *rec;
#ifndef symb_dump_off
  dump_symTab();
#endif
  idx=decls;
  while (idx!=NULL)
    {
      if (idx->entity!=declTag)
	{
	  printf("Error, non decl list passed for popscope.\n");
	  globerr=-1;
	}
      else
	{
	  found=search(idx->u.decl.id);
	  if (found==NULL)
	    {
	      printf("Error, decl not found in symbol table\n");
	      globerr=-1;
	    }
	  else
	    {
	      rec=found->get();
	      if (rec->levelno==levelno)
		{
		  rec->top->r=rec->prev;
		  free(rec->range);
		  if (rec->args!=NULL)
		    {
		      free(rec->args->id);
		      free(rec->args->range);
		      free(rec->args);
		    }
		  free(rec->value);
		  free(rec);
		}
	    }
	}
      idx=idx->next;
    }
  levelno-=lev;
}

/*************************************************************************
Miscellaneous functions
*************************************************************************/

void prtexpand(int typ)
/*************************************************************************
Prints the human readable representation of the type.
*************************************************************************/
{
  if (typ==-1)
    {
      printf("INVALID");
      globerr=-1;
    }
  else
    {
      switch (typ&taBOOLE)
	{
	case tINT:
	  printf("INT");
	  break;
	case tREAL:
	  printf("FLOAT");
	  break;
	case tSTRING:
	  printf("STRING");
	  break;
	case tBOOLE:
	  printf("BOOLE");
	  break;
	case taINT:
	  printf("INTarray");
	  break;
	case taREAL:
	  printf("FLOATarray");
	  break;
	case taSTRING:
	  printf("STRINGarray");
	  break;
	case taBOOLE:
	  printf("BOOLEarray");
	  break;
	}
    }
}

void boolecast(int *op)
/*************************************************************************
Cast different types as BOOLE's
*************************************************************************/
{
  switch(*op&taBOOLE)
    {
    case tINT:
      *op=tBOOLE;
      break;
    case tREAL:
      *op=tBOOLE;
      break;
    case tSTRING:
      *op=tBOOLE;
      break;
    case taINT:
      *op=taBOOLE;
      break;
    case taREAL:
      *op=taBOOLE;
      break;
    case taSTRING:
      *op=taBOOLE;
      break;
    }
}

int cast(int typ1, int typ2, int err, int lineno)
{
      if ((err==0)&&((typ2&taBOOLE)!=(typ1&taBOOLE)))
	{
	  switch(typ2&taBOOLE)
	    {
	    case tBOOLE:
	      err=-1;
	      break;
	    case tSTRING:
	      switch(typ1&taBOOLE)
		{
		case tINT:
		case tREAL:
		  err=-1;
		  break;
		}
	      break;
	    case tREAL:
	      switch (typ1&taBOOLE)
		{
		case tSTRING:
		  err=-1;
		  break;
		}
	      break;
	    case tINT:
	      switch (typ1&taBOOLE)
		{
		case tSTRING:
		  err=-1;
		  break;
		}
	      break;
	    }
	  if (err==-1)
	    {
	      printf("Error[%d], cannot cast a [",lineno);
	      prtexpand(typ2);
	      printf("] into a type [");
	      prtexpand(typ1);
	      printf("]\n");
	      globerr=-1;
	    }
	}
      return err;
}

void incrlabel()
{
  if(currlabel[4]<'9')
    {
      currlabel[4]++;
    }
  else if(currlabel[3]<'9')
    {
      currlabel[3]++;
      currlabel[4]='0';
    }
  else if(currlabel[2]<'9')
    {
      currlabel[2]++;
      currlabel[3]='0';
      currlabel[4]='0';
    }
  else
    {
      printf("Error! Not enough labels\n");
      globerr=-1;
    }
}

void addstr(char **old,char *newst)
{
    char *retval;
    int idx, idx2;
    int size;
    if (newst==NULL)
      return;
    if (*old==NULL)
        size=strlen(newst)+1;
    else
        size=strlen(*old)+strlen(newst)+1;
    retval=(char *)malloc(size);
    for(idx=0;idx<size-strlen(newst)-1;idx++)
        retval[idx]=(*old)[idx];
    for(idx2=0;idx2<strlen(newst);idx2++)
      retval[idx++]=newst[idx2];
    retval[size-1]=0;
    free(*old);
    *old=retval;
}

#endif




