/***************************************************************
tab.h Symbol Table functions header
Handles declarations of identifiers and grows link list of decs
***************************************************************/

#ifndef tab_h
#define tab_h

#include "include.h"
#include <string.h>
#include "generate.h"

HashTab htab;

int varaccess(char *id, A_node *node=NULL, rettype **val=NULL, int ln=0,int *dim=NULL,int **range=NULL, int *loc=NULL);
extern int expreval(A_node expr, rettype **val);
extern int exprevalparam(A_node expr,int *dim, int **range);

void getrange(A_node intyp, int *low, int *high, int dimension=0, int lineno=0)
/*************************************************************************
This function checks the range of a array declaration as well as returns
the upper and lower bounds of the array.  It will do so for the dimension
specified.
*************************************************************************/
{
  int i1;
  A_node typ;
  rettype *val;
  typ=intyp;
  if (lineno==0) lineno=intyp->lineno;

  while (dimension>0)
    { /* Advance to correct dimension */
      if (typ->next==NULL)
	{
	  *low=0;
	  *high=0;
	  return;
	}
      else
	{
	  typ=typ->next;
	  dimension--;
	}
    }/* End while loop */

  /* figure what kind of range specification is used */
  switch(typ->u.typ.kind)
    {
    case array1Ty: /* Specified by size */
      switch(typ->u.typ.base1)
	{
	case inttyp:
	  *high=typ->u.typ.r1.num-1;
	  break;
	case consttyp:
	  i1=varaccess(typ->u.typ.r1.id,NULL,&val,lineno);
	  /* printf("%d\n",i1);*/
	  if((i1&32)==32)
	    {
	      if(val->kind==intTy)
		*high=val->u.intval-1;
	      else
		{
		  printf("Error[%d], invalid range type\n",lineno);
		  globerr=-1;
		}
	    }
	  else
	    {
	      printf("Error[%d], invalid range type\n",lineno);
	      globerr=-1;
	    }
	  break;
	}
      *low=0;
      if(*high < 0)
	{
	  printf("Error[%d], index out of range\n",lineno);
	  globerr=-1;
	}
      break;
    case array2Ty: /* Specified by range */
      switch(typ->u.typ.base1)
	{
	case inttyp:
	  *low=typ->u.typ.r1.num;
	  break;
	case subinttyp:
	  *low=-typ->u.typ.r1.num;
	  break;
	case consttyp:
	  i1=varaccess(typ->u.typ.r1.id,NULL,&val,lineno);

	  if(i1&32==32)
	    {
	      if(val->kind==intTy)
		*low=val->u.intval;
	      else
		{
		  globerr=-1;
		  printf("Error[%d], invalid range type\n",lineno);
		}
	    }
	  else
	    {
	      printf("Error[%d], invalid range type\n",lineno);
	      globerr=-1;
	    }
	  break;
	}
      switch(typ->u.typ.base2)
	{
	case inttyp:
	  *high=typ->u.typ.r2.num;
	  break;
	case subinttyp:
	  *high=-typ->u.typ.r2.num;
	  break;
	case consttyp:
	  i1=varaccess(typ->u.typ.r2.id,NULL,&val,lineno);
	  /*printf("%x\n",i1);*/

	  if(i1&32==32)
	    {
	      if(val->kind==intTy)
		*high=val->u.intval;
	      else
		{
		  globerr=-1;
		  printf("Error[%d], invalid range type\n",lineno);
		}
	    }
	  else
	    {
	      globerr=-1;
	      printf("Error[%d], invalid range type\n",lineno);
	    }
	  break;
	}
      if((*high)<(*low)) /* Check range */
	{
	  printf("Error[%d], invalid range\n",lineno);
	  globerr=-1;
	}
      break;
    }
}

int getdim(A_node typ)
{
  A_node index;
  int counter=0;
  index=typ;
  while(index!=NULL)
    {
      counter++;
      index=index->next;
    }
  return counter;
}


void chkbounds(A_node prg,A_node typ,int idx)
{
  rettype *val;
  int low, high,i;

  i=expreval(prg->u.var.idxlist.expr,&val);
  /*  i=varaccess(prg->u.var.idxlist.u.var->u.var.id,NULL,&val,prg->lineno);*/
  if((i&tCONST)==tCONST)
    { /* It is a constant */
      switch(i&taBOOLE)
	{
	case tINT:
	  getrange(typ,&low,&high,idx,prg->lineno);
	  if((low<=(val)->u.intval) &&(high>=(val)->u.intval))
	    {
	    }
	  else
	    {
	      printf("Error[%d], index out of range\n",prg->lineno);
	      globerr=-1;
	    }
	  break;
	default:
	  globerr=-1;
	  printf("Error[%d], invalid index type\n",prg->lineno);
	  break;
	}
    }
  else
    {
      if ((i&taBOOLE)!=tINT)
	{
	  globerr=-1;
	  printf("Error[%d], invalid index type\n",prg->lineno);
	}
    }
}

void chkarry(A_node prg, A_node typ)
/*************************************************************************
Function checks to see if the array references is valid
eg.  A[3,8]  is valid if var A:array [2..5,9] of base;
A[1,5] A[4,9] A[-1,-1] are wrong.
*************************************************************************/
{
  A_node idx;
  int dim,i=0;
  int *range;
  varaccess(prg->u.var.id,NULL,NULL,prg->lineno,&dim,&range);
  idx=prg;
  while(idx!=NULL)
    {
      if(i<=dim)
	chkbounds(idx,typ,i);
      i++;
      idx=idx->u.var.idxlist.next;
    }
  if(dim>i)
    {
      globerr=-1;
      printf("Error[%d], not enough dimensions specified\n",prg->lineno);
    }
  else if (dim<i)
    {
      printf("Error[%d], too many dimensions specified\n",prg->lineno);
      globerr=-1;
    }
}

int drmatch(int d1,int *r1,int d2,int *r2,int line, char *id1, char *id2)
{
  int idx;
  if(d1!=d2)
    {
      printf("Error[%d], [%s] dimensions do not match [%s]\n",line,id1,id2);
      globerr=-1;
      return -1;
    }
  else
    {
      for (idx=0;idx<d1;idx++)
	{
	  if(r1[idx]!=r2[idx])
	    {
	      printf("Error[%d], range for [%s],[%s] in dimension %d does not match\n",line,id1,id2,idx+1);
	      globerr=-1;
	      return -1;
	    }
	}
    }
  return 0;
}

int dimmatch(A_node var, A_node expr,int *dim2, int **range2, char *id1, char *id2)
{
  int dim, *range,i;
  varaccess(var->u.var.id,0,0,0,&dim,&range);
  if((dim2==NULL)||(range2==NULL))
    {
      return -1;
    }
  i=drmatch(dim,range,*dim2,*range2,var->lineno,id1,id2);
  return i;
}

void update(char *funcid,char *varid,int byref,int typ, int dim, int **range)
{
  SYMTAB_ENTRY *ret;
  DRlist *entry, *idx;

  ret=htab.search(funcid);
  if (ret->r->args==NULL)
    {
      entry=(DRlist *)malloc(sizeof(DRlist));
      entry->id=(char *)malloc(strlen(varid));
      strcpy(entry->id,varid);
      entry->typ=typ;
      entry->dim=dim;
      entry->range=*range;
      entry->next=NULL;
      ret->r->args=entry;
      //initialize
    }
  else
    {
      idx=ret->r->args;
      while (idx->next!=NULL)
	{
	  idx=idx->next;
	}
      entry=(DRlist *)malloc(sizeof(DRlist));
      entry->id=(char *)malloc(strlen(varid));
      strcpy(entry->id,varid);
      entry->typ=typ;
      entry->dim=dim;
      entry->range=*range;
      entry->next=NULL;
      idx->next=entry;
      //just add param in arg list.
    }

  return;
}

int match(DRlist *form,A_node act)
{
  int typ,dim,ret=0;
  int *range;
  
  typ=exprevalparam(act,&dim,&range);
  if (form->typ!=typ)
    {
      ret=cast(form->typ,typ,ret,act->lineno);
      if (((form->typ&tBOOLE)!=(typ&tBOOLE))&&(((form->typ&taINT)==taINT)&&((typ&taINT)==taINT)))
	{
	  printf("Error[%d], cannot cast array types\n",act->lineno);
	  globerr=-1;
	  ret=-1;
	}
    }
  if ((ret!=-1) && ((form->typ&taINT)==taINT))
    {
      if (drmatch(form->dim,form->range,dim,range,act->lineno,"operand1","operand2")==-1)
	ret=-1; 
    }
  free(range);
  return ret;
}

void checksubcall(char *funcid,A_node exprlst)
{
  SYMTAB_ENTRY *ret;
  A_node idx2;
  DRlist *entry, *idx;
  int i;

  ret=htab.search(funcid);
  gensubcall1(symtable->namestr,funcid);
  if (ret!=NULL)
    {
      if (ret->r->args==NULL)
	{
	  if (exprlst!=NULL)
	    {
	      printf("Error[%d], no arguments specified in formal declaration\n",exprlst->lineno);
	      globerr=-1;
	    }
	}
      else
	{
	  idx=ret->r->args;
	  idx2=exprlst;
	  i=1;
	  while ((idx!=NULL)&&(idx2!=NULL))
	    {
	      if (match(idx,idx2)==-1)
		{
		  printf("Error[%d], actual argument type does not match formal type for parameter [%d]\n",idx2->lineno,i);
		  globerr=-1;
		}
	      gensubcallmid(idx->typ,idx->dim);
	      i++;
	      idx=idx->next;
	      idx2=idx2->next;
	    }
	  gensubcall2(ret->r->typ);
	  if (idx!=NULL)
	    {
	      printf("Error[%d], too many arguments specified in formal declaration [%s]\n",exprlst->lineno,funcid);
	      globerr=-1;
	    }
	  if (idx2!=NULL)
	    {
	      printf("Error[%d], too few arguments specified in formal declaration [%s]\n",exprlst->lineno,funcid);
	      globerr=-1;
	    }
	}
    }
}

void makerec(int typ, A_node dec, SYMTAB_ENTRY *ret, int flag,rettype *val, int loc)
{
  int i, low, high;
  SYMTAB_REC *rec;
  rec=(SYMTAB_REC *)malloc(sizeof(SYMTAB_REC));
  rec->top=ret;
  rec->typ=typ;
  if (flag==1)
    rec->prev=ret->r;
  else
    rec->prev=NULL;
  rec->decl=dec;
  rec->levelno=levelno;
  rec->value=val;
  if(levelno>1)
    rec->sloc=loc;
  else
    rec->sloc=-1;
  rec->args=NULL;
  if((dec!=NULL) && ((typ&taINT)==taINT))
    {
      rec->dim=getdim(dec);
      rec->range=(int *)malloc(rec->dim*sizeof(int));
      for(i=0;i<rec->dim;i++)
	{
	  getrange(dec,&low,&high,i,dec->lineno);
	  rec->range[i]=high-low+1;
	}
    }
  else
    {
      rec->dim=0;
      rec->range=NULL;
    }
  ret->r=rec;
}

int decl(char *id, A_node dec, A_node *list, int typ, rettype *val=NULL, int ln=0, int loc=-1)
/*************************************************************
Declare a id which came from dec and append to link list list
*************************************************************/
{
  int line;
  A_node idx,lst;
  SYMTAB_ENTRY *ret;
  SYMTAB_REC *rec;
  char *name;
  int i;
  
  if (dec!=NULL)
    line=dec->lineno;
  else
    line=ln;
  
  lst=*list;
  ret=htab.search(id);
  if (ret==NULL)
    {    /* Identifier does not yet exist */
      if ((symtabPtr-symtable)>1024)
	{
	  printf("Error[%d], maximum symbols exceeded.\n",line);
	  globerr=-1;
	  return 1;
	}
      ret=symtabPtr;
      symtabPtr++;
      name=(char *)malloc(strlen(id)+1);
      strcpy(name,id);
      
      makerec(typ, dec,ret,0,val,loc);
      
      ret->namestr=name;
      ret->put(levelno);
      
      idx=(A_node)malloc(sizeof(A_Node));
      idx->next=lst;
      idx->entity=declTag;
      idx->lineno=0;
      idx->u.decl.id=id;
      (*list)=idx;
    }
  else
    {   /* Identifier exists but out of scope */
      if (ret->get()==NULL)
	{
	  makerec(typ,dec,ret,0,val,loc);
	  
	  idx=(A_node)malloc(sizeof(A_Node));
	  idx->next=lst;
	  idx->entity=declTag;
	  idx->lineno=0;
	  idx->u.decl.id=id;
	  (*list)=idx;
	  return 0;
	}
      /* printf("%s[%d]:[%d]\n",ret->namestr,ret->get()->levelno,levelno);/**/
      if (ret->get()->levelno==levelno)
	{     /* Identifier already declared in this scope */
	  printf("Error[%d], duplicate declaration of %s\n",line,ret->namestr);
	  globerr=-1;
	  return 1;
	}
      
      /* Identifier only declared in higher scope, redeclare */
      makerec(typ,dec,ret,1,val,loc);
      
      idx=(A_node)malloc(sizeof(A_Node));
      idx->entity=declTag;
      idx->u.decl.id=id;
      idx->next=lst;
      (*list)=idx;
    }
  return 0;
}

int varaccess(char *id, A_node *node=NULL, rettype **val=NULL, int ln=0,int *dim=NULL,int **range=NULL,int *loc=NULL)
/*****************************************************************
Check if id has already been declared within the scope of this
statement.
*****************************************************************/
{
  int line=0;
  int i;
  SYMTAB_ENTRY *ret;
  /*printf("%s\n",id);/**/
  ret=htab.search(id);
  if ((ret==NULL) || (ret->get()==NULL))
    {
      line=ln;
      printf("Error[%d], %s is not declared.\n",line,id);
      globerr=-1;
      return -1;
    }
  if (node!=NULL)
    {
      *node=ret->get()->decl;
    }
  if(val!=NULL)
    {
      *val=ret->get()->value;
    }
  if(dim!=NULL)
    {
      *dim=ret->get()->dim;
    }
  if(range!=NULL)
    {
      *range=ret->get()->range;
    }
  if(loc!=NULL)
    {
      *loc=ret->get()->sloc;
    }
  i=ret->get()->typ;
  /*  printf("i=%d\n",i);*/
  return i;
}

int getlocation(char *id)
{
  SYMTAB_ENTRY *ret;
  ret=htab.search(id);
  return ret->get()->sloc;
}


#endif


